% % The main function of the code is to generate and save conductivity distributions and their corresponding boundary voltage data for electrical impedance tomography (EIT) modeling. The flow of the code is roughly as follows:
% % % 
% % 1. **Create directory**: Create storage directories for different types of data (e.g., boundary voltages with different signal-to-noise ratios, conductivity, etc.).
% % 2. **Model Definition**: Define a 3D cylindrical EIT model including its dimensions, number of electrodes, electrode positions, etc. % % 3.
% % 3. **Conductivity Distribution Generation**: Generate conductivity values for each element in the model based on a randomly selected center point and a specified range of decay rates.
% % 4. **Calculate Boundary Voltage**: Calculates the boundary voltage for the defined model.
% % 5. **Add Noise**: Adds different levels of noise to the calculated boundary voltage.
% % 6. **Save Data**: Saves the generated conductivity distribution, the boundary voltage without noise and the boundary voltage with noise to the appropriate directory.
% % 7. **Visualization**: Visualize the generated conductivity model and save it as a .fig file using the EIDORS tool.
% % % 
% % The following are the key variables in the code:
% % % 
% % - **Directory variables**: such as `DDL_dir`, `fig_dir`, `BV_dir`, etc., which define where the different data types are stored.
% % - **Model parameters**.
% % - `cyl_shape`: Defines the size of the cylinder model.
% % - `nelec`: Number of electrodes per layer.
% % - `ring_vert_pos`: Position of the electrode ring.
% % - `elec_shape`: The shape of the electrodes.
% % - **Conductivity distribution parameters**.
% % - `base_scale_x/y/z`: Defines the range of base decay rates of the conductivity in each direction.
% % - `diffusion_range`: The range of random perturbations in the decay rate of the conductivity.
% % - `center_position_perturbation`: Random perturbation range for the center position.
% % - `distance_perturbation_scale`: Random perturbation range for distance calculation.
% % - `min/max_conductivity_range`: Minimum and maximum range of conductivity values.
% % - `num_centers_z/xy_range`: Range of the number of centers in the Z-direction and XY-plane.
% % - `img1`: EIDORS image object storing model and conductivity data.
% % - `vi`: Boundary voltage data without noise.
% % - `vi_10dB`, `vi_20dB` etc.: Boundary voltage data with different signal-to-noise ratios.
% % % 
% % These variables and parameters provide you with complete control over the EIT model conductivity distribution and boundary voltage data generation process. 
% You can adjust these parameters as needed to meet your specific application requirements.*** Translated with www.DeepL.com/Translator (free version) ***

mkdir absolute_maize_2W_95295_by_nodes nodes
mkdir absolute_maize_2W_95295_by_nodes DDL;
mkdir absolute_maize_2W_95295_by_nodes fig;
mkdir absolute_maize_2W_95295_by_nodes BV;
mkdir absolute_maize_2W_95295_by_nodes BV_reshape;
mkdir absolute_maize_2W_95295_by_nodes\BV_10dB;
mkdir absolute_maize_2W_95295_by_nodes\BV_10dB_reshape;
mkdir absolute_maize_2W_95295_by_nodes\BV_20dB;
mkdir absolute_maize_2W_95295_by_nodes\BV_20dB_reshape;
mkdir absolute_maize_2W_95295_by_nodes\BV_30dB;
mkdir absolute_maize_2W_95295_by_nodes\BV_30dB_reshape;
mkdir absolute_maize_2W_95295_by_nodes\BV_40dB;
mkdir absolute_maize_2W_95295_by_nodes\BV_40dB_reshape;
mkdir absolute_maize_2W_95295_by_nodes\BV_50dB;
mkdir absolute_maize_2W_95295_by_nodes\BV_50dB_reshape;

nodes_dir = './absolute_maize_2W_95295_by_nodes/nodes/';
DDL_dir = './absolute_maize_2W_95295_by_nodes/DDL/';
fig_dir = './absolute_maize_2W_95295_by_nodes/fig/';
BV_dir = './absolute_maize_2W_95295_by_nodes/BV/';
BV_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_reshape/';
BV_10dB_dir = './absolute_maize_2W_95295_by_nodes/BV_10dB/';
BV_10dB_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_10dB_reshape/';
BV_20dB_dir = './absolute_maize_2W_95295_by_nodes/BV_20dB/';
BV_20dB_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_20dB_reshape/';
BV_30dB_dir = './absolute_maize_2W_95295_by_nodes/BV_30dB/';
BV_30dB_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_30dB_reshape/';
BV_40dB_dir = './absolute_maize_2W_95295_by_nodes/BV_40dB/';
BV_40dB_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_40dB_reshape/';
BV_50dB_dir = './absolute_maize_2W_95295_by_nodes/BV_50dB/';
BV_50dB_reshape_dir = './absolute_maize_2W_95295_by_nodes/BV_50dB_reshape/';

num = 5; 

for k = 1:num
    cyl_shape = [0.066,0.0355,0.003]; 
    nelec= 16; 
    ring_vert_pos = [0.011, 0.033, 0.055]; 
    nrings= 3; 
    elec_shape = [0.001,0,0.02]; 

    fmdl= ng_mk_cyl_models(cyl_shape,[nelec,ring_vert_pos],elec_shape); 
    
    imdl = mk_common_model('b3cr',[16,3]); % Will replace most fields
    imdl.fwd_model = fmdl;
    imdl.fwd_model.stimulation = mk_stim_patterns(nelec,nrings,[0,1],[0,1],{'no_meas_current','rotate_meas'},0.03); % 10mA
    img1 = mk_image(imdl, 1); 
    show_fem(img1)
    show_3d_slices(img1,0,[],0)
    

%% Distribution of Electrical Conductivity Inside Corn ear
%------------------------- Approach 3----------------------------------------
% ----------- Customizable Variables -------------
% The goal of this code is to generate a distribution of electrical conductivity within a 3D model. Here's an explanation of the basic principles and key steps:
%
% 1. Parameter Definitions:
% - base_scale_x/y/z: These parameters define the attenuation rate range in each direction. The attenuation rate determines how electrical conductivity decreases with increasing distance.
% - diffusion_range: This is a factor used to randomly adjust the attenuation rate, introducing some randomness into the electrical conductivity distribution in the model.
% - center_position_perturbation: Defines the range of random perturbation for the center point position, increasing uncertainty in the center position.
% - distance_perturbation_scale: Adds random perturbation to the calculated distances for increased variability in distances from each element to the center point.
% - min/max_conductivity_range: Defines the minimum and maximum range of electrical conductivity.
% - num_centers_z_range/num_centers_xy_range: Defines the number of centers in the Z direction and XY plane.
%
% 2. Calculate Center Coordinates for All Elements:
% - Calculate the center coordinates for each element using the model's node and element information.
%
% 3. Determine the Boundaries of the 3D Field:
% - Find the minimum and maximum coordinates of the centers of all elements in the model.
%
% 4. Randomly Select Centers in the Z Direction:
% - Randomly select centers in the Z direction.
%
% 5. Initialize Electrical Conductivity:
% - Initialize an electrical conductivity value for each element, initially set to zero.
%
% 6. For Each Center in the Z Direction:
% - Enter a loop to operate on each center in the Z direction.
%
% 7. Select Centers in the XY Direction:
% - For each center in the Z direction, enter another loop to select centers in the XY direction.
%
% 8. Calculate Distances Between Elements and Centers:
% - Calculate the distance from the center of each element to the current center in the XY and Z directions.
%
% 9. Determine Diffusion Rate:
% - Randomly select an attenuation rate within the specified range, considering diffusion_range for random perturbation.
%
% 10. Calculate the Effect on Electrical Conductivity:
% - Use an exponential function to simulate how electrical conductivity changes with distance. This involves the attenuation rate and the distance between elements and centers.
%
% 11. Normalize and Map Electrical Conductivity:
% - Use normalization to map the range of electrical conductivity to the specified minimum and maximum conductivity range.
%
% 12. Output the Results:
% - Assign the computed electrical conductivity distribution to img1.elem_data.
%
% In summary, the goal of this code is to generate an electrical conductivity value for each element in a 3D model. 
% These values are calculated based on randomly selected centers, specified attenuation rate ranges, and other parameters. 
% The calculation of electrical conductivity takes into account factors such as distance between elements and centers, random perturbations, and more.
    
    base_scale_x = [1, 5]; % Range of decay speeds in the X-direction
    base_scale_y = [1, 5]; % Range of decay speed in the Y-direction
    base_scale_z = [4, 9]; % Range of decay speeds in the Z-direction
    
    diffusion_range = [0.5, 1.5]; % Random range of diffusion rates
    
    center_position_perturbation = [0.01, 0.2]; 
    distance_perturbation_scale = [0.001, 0.01]; 
    
    min_conductivity_range = [1, 2.5]; 
    max_conductivity_range = [3.5, 4.5]; 
    
    num_centers_z_range = [5, 20]; 
    num_centers_xy_range = 20; 

    elem_centroids = (fmdl.nodes(fmdl.elems(:,1), :) + fmdl.nodes(fmdl.elems(:,2), :) + fmdl.nodes(fmdl.elems(:,3), :) + fmdl.nodes(fmdl.elems(:,4), :)) / 4;

    min_coords = min(elem_centroids);
    max_coords = max(elem_centroids);
    range_coords = max_coords - min_coords;

    center_z = min_coords(3) + rand(1, randi(num_centers_z_range)) * range_coords(3);

    elems = zeros(size(elem_centroids, 1), 1);

    for i = 1:length(center_z)
        for j = 1:randi(num_centers_xy_range)
            center_xy = [min_coords(1) + (0.4 + (center_position_perturbation(2)-center_position_perturbation(1))*rand() + center_position_perturbation(1)) * range_coords(1), 
                         min_coords(2) + (0.4 + (center_position_perturbation(2)-center_position_perturbation(1))*rand() + center_position_perturbation(1)) * range_coords(2)];
            center_xy = center_xy';
            distances_xy = sqrt(sum((elem_centroids(:,1:2) - center_xy).^2, 2)) + (distance_perturbation_scale(2)-distance_perturbation_scale(1))*randn(size(elem_centroids, 1), 1) + distance_perturbation_scale(1);
            distance_z = abs(elem_centroids(:,3) - center_z(i));
            scale_x = (base_scale_x(1) + rand() * (base_scale_x(2) - base_scale_x(1))) * (diffusion_range(1) + (diffusion_range(2)-diffusion_range(1))*rand());
            scale_y = (base_scale_y(1) + rand() * (base_scale_y(2) - base_scale_y(1))) * (diffusion_range(1) + (diffusion_range(2)-diffusion_range(1))*rand());
            scale_z = (base_scale_z(1) + rand() * (base_scale_z(2) - base_scale_z(1))) * (diffusion_range(1) + (diffusion_range(2)-diffusion_range(1))*rand());
            
            scale = [scale_x, scale_y, scale_z];
            elem_effect = exp(-distances_xy / scale(1)) .* exp(-distances_xy / scale(2)) .* exp(-distance_z / scale(3));
            elems = elems + elem_effect;
        end
    end
    normalized_elems = (elems - min(elems)) / (max(elems) - min(elems));
    min_conductivity = min_conductivity_range(1) + rand() * (min_conductivity_range(2) - min_conductivity_range(1));
    max_conductivity = max_conductivity_range(1) + rand() * (max_conductivity_range(2) - max_conductivity_range(1));
    elems = min_conductivity + normalized_elems * (max_conductivity - min_conductivity);
    img1.elem_data = elems;
        writematrix(elem_centroids, strcat(nodes_dir,string(k),'.csv'));

        vi=fwd_solve(img1);    
        writematrix(vi.meas,strcat(BV_dir,string(k),'.csv'));
        vi_reshape = reshape(vi.meas,[45,48]);
        writematrix(vi_reshape,strcat(BV_reshape_dir,string(k),'.csv'));
        writematrix(img1.elem_data,strcat(DDL_dir,string(k),'_ddl.csv')); 
        h1=subplot(1,2,1);
        show_fem(img1);
        h2=subplot(1,2,2);
        show_3d_slices(img1,[0.011, 0.033, 0.055], 0,0);
        axis tight; axis equal;
        common_colourbar([h1,h2],img1); 
        colormap(jet);
        filename = strcat(fig_dir,string(k),'.fig');
        saveas(gcf,filename);
    
    %     SNRdB = 20;
        vi_10dB = add_noise(10^(10/20) ,vi);
        writematrix(vi_10dB.meas,strcat(BV_10dB_dir,string(k),'.csv'));
        vi_10dB_reshape = reshape(vi_10dB.meas,[45,48]);
        writematrix(vi_10dB_reshape,strcat(BV_10dB_reshape_dir,string(k),'.csv'));
    %     SNRdB = 20;
        vi_20dB = add_noise(10^(20/20) ,vi);
        writematrix(vi_20dB.meas,strcat(BV_20dB_dir,string(k),'.csv'));
        vi_20dB_reshape = reshape(vi_20dB.meas,[45,48]);
        writematrix(vi_20dB_reshape,strcat(BV_20dB_reshape_dir,string(k),'.csv'));
    
    %     SNRdB = 30;
        vi_30dB = add_noise(10^(30/20) ,vi);
        writematrix(vi_30dB.meas,strcat(BV_30dB_dir,string(k),'.csv'));
        vi_30dB_reshape = reshape(vi_30dB.meas,[45,48]);
        writematrix(vi_30dB_reshape,strcat(BV_30dB_reshape_dir,string(k),'.csv'));
    
    %     SNRdB = 40; 
        vi_40dB = add_noise(10^(40/20) ,vi);
        writematrix(vi_40dB.meas,strcat(BV_40dB_dir,string(k),'.csv'));
        vi_40dB_reshape = reshape(vi_40dB.meas,[45,48]);
        writematrix(vi_40dB_reshape,strcat(BV_40dB_reshape_dir,string(k),'.csv'));
    
        %     SNRdB = 50; 
        vi_50dB = add_noise(10^(50/20) ,vi);
        writematrix(vi_50dB.meas,strcat(BV_50dB_dir,string(k),'.csv'));
        vi_50dB_reshape = reshape(vi_50dB.meas,[45,48]);
        writematrix(vi_50dB_reshape,strcat(BV_50dB_reshape_dir,string(k),'.csv'));
    
    % %     SNRdB = 80; 
    %     vi_80dB = add_noise(10^( 80/20) ,vi);
    %     writematrix(vi_80dB.meas,strcat(BV_80dB_dir,string(k),'.csv'));
    %     vi_80dB_reshape = reshape(vi_80dB.meas,[45,48]);
    %     writematrix(vi_80dB_reshape,strcat(BV_80dB_reshape_dir,string(k),'.csv'));

end
