% Visual refactoring
% Use uigetfile to select one or more .csv files
[filename, pathname] = uigetfile('E:\*.csv', 'Select the .csv file to be processed', 'MultiSelect', 'on');
if isequal(filename,0)
   disp('User Selection Cancellation')
   return;
else
   if iscell(filename) 
       fileall = strcat(pathname, filename);
   else
       fileall = {fullfile(pathname, filename)}; 
   end
   disp(['The user has selected the following file：', strjoin(fileall, ', ')])
end

save_fig = uigetdir('E:\', 'Select Save Address');
if isequal(save_fig,0)
   disp('User Selection Cancellation')
   return;
else
   disp(['The file will be saved to the', save_fig])
end

for i = 1:length(fileall)
    figure;  

    dir = fileall{i};

    cyl_shape = [0.066,0.0355,0.003]; 
    nelec= 16; 
    ring_vert_pos = [0.011, 0.033, 0.055]; 
    nrings= 3; 
    elec_shape = [0.001,0,0.02]; 
    fmdl= ng_mk_cyl_models(cyl_shape,[nelec,ring_vert_pos],elec_shape); 

    imdl = mk_common_model('b3cr',[16,3]); % Will replace most fields
    imdl.fwd_model = fmdl;
    imdl.fwd_model.stimulation = mk_stim_patterns(nelec,nrings,[0,1],[0,1],{'no_meas_current','rotate_meas'},0.03); % 10mA
    img1 = mk_image(imdl, 0.035);  

    sig3 = readmatrix(dir);
    img1.elem_data = sig3;

        %% Customized 3D whole map and slice map visualization
    
    nodes = img1.fwd_model.nodes;
    elems = img1.fwd_model.elems;
    
    elec_data = img1.elem_data;
    
    min_elec_data = min(elec_data);
    max_elec_data = max(elec_data);
    
    elem_centroids = (nodes(elems(:,1), :) + nodes(elems(:,2), :) + nodes(elems(:,3), :) + nodes(elems(:,4), :)) / 4;

    buffer = 0.01; 
    resolution = 150; % Increased resolution
    [x, y, z] = meshgrid(linspace(min(elem_centroids(:,1)) - buffer, max(elem_centroids(:,1)) + buffer, resolution), ...
                         linspace(min(elem_centroids(:,2)) - buffer, max(elem_centroids(:,2)) + buffer, resolution), ...
                         linspace(min(elem_centroids(:,3)) - buffer, max(elem_centroids(:,3)) + buffer, resolution));
    
    % Use 'linear' interpolation
    V = griddata(elem_centroids(:,1), elem_centroids(:,2), elem_centroids(:,3), elec_data, x, y, z, 'linear');
    
    % Create graphs and subgraphs
    disp(['Processing file ', num2str(i), ' of ', num2str(length(fileall))]);
    
    % Subfigure 1: 3D overall view
    subplot(1,2,1);
    shading interp;
    material shiny;
    colormap jet;
    % Add all four faces to the list of triangles
    tris = [elems(:,[1,2,3]);
            elems(:,[1,2,4]);
            elems(:,[1,3,4]);
            elems(:,[2,3,4])];
    
    % Repeat the conductivity data four times to match the new list of triangles
    elec_data_repeated = repmat(elec_data, 4, 1);
    
    h = trisurf(tris, nodes(:,1), nodes(:,2), nodes(:,3), elec_data_repeated, 'EdgeColor', 'none');
    colorbar;
    caxis([min_elec_data, max_elec_data]);
    axis equal tight;
    view(3);
    xlabel('X');
    ylabel('Y');
    zlabel('Z');
    title('3D Conductivity Visualization with trisurf');
    
    subplot(1,2,2);
    slice_heights = [0.011, 0.033, 0.055];
    for j = 1:length(slice_heights)
        h = slice(x, y, z, V, [], [], slice_heights(j));
        set(h,'EdgeColor','none', 'FaceColor','interp');
        alpha(0.8);
        hold on;
    end
    h = slice(x, y, z, V, 0, [], []);
    set(h,'EdgeColor','none', 'FaceColor','interp');
    alpha(0.8);
    hold on;
    h = slice(x, y, z, V, [], 0, []);
    set(h,'EdgeColor','none', 'FaceColor','interp');
    alpha(0.8);
    colorbar;
    caxis([min_elec_data, max_elec_data]);
    colormap(jet);
    title('3D Conductivity Slices at specified heights and at X, Y origins');
    xlabel('X'); ylabel('Y'); zlabel('Z');
    axis tight;
    axis equal;
    view(3);

    if iscell(filename)
        current_filename = filename{i};
    else
        current_filename = filename;
    end

    saveas(gcf, fullfile(save_fig, [current_filename, '.fig']))
    saveas(gcf, fullfile(save_fig, [current_filename, '.png']))
    close(gcf);  

end
