# -*- coding: utf-8 -*-
""" 
@Time    : 2022/12/28 0:10
@Author  : hying
@Emily   : haiying_ciee@cau.edu.cn
"""

import argparse
import os
import random
import numpy
import time
import datetime
from custom_function.function import CarvanaDataset, BasicDataset, \
    Summary, AverageMeter, ProgressMeter, save_checkpoint
from custom_function.function import Logger
from Network import ResNet, Vgg16, UNet
from torchsummary import summary
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import wandb
import sys
from tqdm import tqdm
# from pytorchtools import EarlyStopping


def main(datadir, main_bv, main_style):
    global es, optimizer, scheduler
    torch.cuda.empty_cache()
    best_RMSE = 100000
    args = parser.parse_args()

    if args.wandb is True:
        wandb.init(
            project=args.Net,
            resume='allow',
            entity='eit'
        )
    save_time = datetime.datetime.now().strftime('%m-%d-%H-%M')

    '''
        Setup the device(GPU or CPU)
    '''

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu), '; The name of device:',
              torch.cuda.get_device_name(args.gpu))
        device = torch.device(f"cuda:{args.gpu}")
    else:

        if torch.cuda.is_available():
            print('Using device GPU',
                  '; The number of device:', torch.cuda.device_count(),
                  '; The name of device:', torch.cuda.get_device_name(0))
            device = torch.device("cuda:0")
        else:
            print('Using device CPU')
            device = torch.device("cpu")


    '''
    Set the random seed, in order to reproduce the experimental results
    '''
    if args.seed is not None:
        random.seed(args.seed)
        numpy.random.seed(args.seed)
        torch.manual_seed(args.seed)
        torch.cuda.manual_seed_all(args.seed)
        cudnn.deterministic = True

    '''
    Create model
    '''
    if args.Net == "ResNet50":
        model = ResNet.resnet50(pretrained=args.pretrained)
        model.conv1 = nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
        model.fc = nn.Linear(model.fc.in_features, args.num_elements)
    elif args.Net =="Vgg16":
        model = Vgg16.Vgg16(in_channels=1, out_classes=95295)
    elif args.Net =="UNet":
        model = UNet.UNet(n_channels=1, n_classes=args.num_elements)
    else:
        raise ValueError(f"Invalid model name '{args.Net}'")
    summary(model.cuda(), input_size=(1, 45, 48))

    model = model.to(device)

    #-------------------------------------------------------
    if args.pretrained is True:
        print('Using pretrained model')
    else:
        print('Do not use pretrained model')

    print(
        'learning rate:', args.lr,
        '\nbatch size:', args.batch_size,
        '\nepoch:', args.epochs
    )
    # --------------------------------------------------------

    '''
    Loss function, optimizer, learning rate decay, mixed precision
    '''
    criterion = nn.MSELoss().to(device)  # move the loss function to the device
    print('Loss function: MSELoss')

    if args.optimizers == 'SGD':
        optimizer = torch.optim.SGD(
            model.parameters(), args.lr,
            momentum=args.momentum,
            weight_decay=args.weight_decay
        )
    elif args.optimizers == 'Adam':
        optimizer = torch.optim.Adam(model.parameters(), args.lr, weight_decay=0.000001, amsgrad=True)
    else:
        raise ValueError(f"Invalid optimizer '{args.optimizers}'")
    print('Optimizer:', args.optimizers)

    # create learning rate scheduler
    if args.schedulers == 'StepLR':
        scheduler = torch.optim.lr_scheduler.StepLR(
            optimizer,
            step_size=args.step_size,  # step_size defult is 30
            gamma=0.1
        )

    elif args.schedulers == 'RLROP':
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', factor=0.1, patience=3, verbose=True,
                                                               min_lr=0.000001)
    else:
        raise ValueError(f"Invalid scheduler '{args.schedulers}'")
    print('Scheduler: {}'.format(args.schedulers))
    # grad_scaler = torch.cuda.amp.GradScaler(enabled=args.amp)

    '''
    Resume from checkpoint if specified
    '''
    if args.resume:
        if os.path.isfile(args.resume):  # Check if the checkpoint file exists
            print("=> loading checkpoint '{}'".format(args.resume))
            loc = 'cuda:{}'.format(
                args.gpu) if args.gpu is not None else 'cpu'  # Specify the location to load the checkpoint
            checkpoint = torch.load(args.resume, map_location=loc)
            args.start_epoch = checkpoint['epoch']
            best_RMSE = checkpoint['best_RMSE']
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            scheduler.load_state_dict(checkpoint['scheduler'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
        else:
            raise FileNotFoundError("No checkpoint found at '{}'".format(args.resume))

    cudnn.benchmark = True

    '''
    Data loading
    '''
    dir_img = os.path.join(datadir, main_bv)
    dir_mask = os.path.join(datadir, args.ddl)
    print("Boundary voltage data:", dir_img)
    print("Conductivity data：", dir_mask)

    dataset = BasicDataset(dir_img, dir_mask)
    print('Sample size of dataset：', len(dataset))

    n_val = int(len(dataset) * args.val_percent)
    n_train = len(dataset) - n_val
    train_set, val_set = torch.utils.data.random_split(
        dataset, [n_train, n_val],
        generator=torch.Generator().manual_seed(0)
    )

    # train dataset
    train_loader = torch.utils.data.DataLoader(
        train_set, batch_size=args.batch_size, shuffle=True,
        num_workers=args.workers, pin_memory=True, sampler=None, persistent_workers=True)

    # val dataset
    val_loader = torch.utils.data.DataLoader(
        val_set, batch_size=args.batch_size, shuffle=False,
        num_workers=args.workers, pin_memory=True)

    '''
   iterative cycle
    '''
    # global glob_step
    glob_step = 0

    for epoch in range(args.start_epoch, args.epochs + 1):
        temp_step = train(train_loader, model, criterion, optimizer, epoch, args, n_train, glob_step)
        loss, RMSE = validate(val_loader, model, criterion, epoch, args, n_val, glob_step)
        scheduler.step(loss)
        print('Average_Loss={}, Average_RMSE={}, [When Epoch={}]'.format(loss, RMSE, epoch))

        glob_step = temp_step

        # Join the early stop
        if RMSE < best_RMSE:
            best_RMSE = min(RMSE, best_RMSE)
            filename = './output/save_model/BestRMSE_{}_model_{}epochs_{}lr_{}.pth'
            torch.save(model, filename.format(main_style, args.epochs, args.lr, save_time))
            print('The optimal model is saved. Epoch is {}'.format(epoch))
            # es = 0
        else:
            print('Not an optimal model')
            # es += 1
            # print("EarlyStopping counter: {} of {}".format(es, args.es_patience))
            # if es >= args.es_patience:
            #     print("Early stopping with best_RMSE: ", best_RMSE, ";and the average val_loss for this epoch: ", loss)
            #     break


def train(train_loader, model, criterion, optimizer, epoch, args, n_train, glob_step):
    model.train()

    with tqdm(total=n_train, desc=f'Epoch[{epoch}/{args.epochs}]', leave=True, unit='csv', unit_scale=True) as pbar:
        for i, (bv, target) in enumerate(train_loader):
            # boundary_voltage = boundary_voltage.cuda(non_blocking=True).to(dtype=torch.float32).squeeze(1)  # RNN
            # bv = torch.reshape(bv, (-1, 1, 45, 48))
            # bv = bv.cuda(non_blocking=True).to(dtype=torch.float32)
            if args.reshape is True:
                bv = torch.reshape(bv, (-1, 1, 45, 48))  # reshape
                bv = bv.cuda(non_blocking=True).to(dtype=torch.float32)
            else:
                bv = bv.cuda(non_blocking=True).to(dtype=torch.float32)
            # ddl = ddl.cuda(non_blocking=True).to(dtype=torch.float32).squeeze(-1).unsqueeze(0)

            target = target.cuda(non_blocking=True).to(dtype=torch.float32).squeeze(-1) * args.ampl
    
            # with torch.cuda.amp.autocast(enabled=args.amp):
            output = model(bv).to(dtype=torch.float32)  # outsize
            loss = criterion(output, target)
            # RMSE_train
            RMSE = torch.sqrt(loss)

            # wandb
            if args.wandb is True:
                wandb.log({
                    'train_loss': loss,
                    'train_RMSE': RMSE,
                    'learning rate': optimizer.param_groups[0]['lr'],
                    'epoch': epoch,
                    'glob_step': glob_step
                })

            # losses.update(loss.item(), bv.size(0))
            # RMSEs.update(RMSE, bv.size(0))

            # Calculate the gradient and perform the SGD step
            optimizer.zero_grad()
            loss.backward()
            # RMSE_train.backward()
            optimizer.step()

            # optimizer.zero_grad(set_to_none=True)
            # grad_scaler.scale(loss).backward()
            # grad_scaler.step(optimizer)
            # grad_scaler.update()

            glob_step += 1
            pbar.set_postfix({'loss': loss.item(), 'RMSE': RMSE.item()})
            pbar.update(bv.size(0))
    return glob_step


def validate(val_loader, model, criterion, epoch, args, n_val, glob_step):
    losses = AverageMeter('Loss', ':.4e', Summary.NONE)
    RMSEs = AverageMeter('RMSE_val', ':6.4f')

    model.eval()
    with torch.no_grad():
        with tqdm(total=n_val, desc=f'Val[{epoch}/{args.epochs}]', leave=True, unit='it', unit_scale=True) as pbar:
            for i, (bv, target) in enumerate(val_loader):
                bv = torch.reshape(bv, (-1, 1, 45, 48))
                bv = bv.cuda(non_blocking=True).to(dtype=torch.float32)
                target = target.cuda(non_blocking=True).to(dtype=torch.float32).squeeze(-1) * args.ampl
                output = model(bv)
                loss = criterion(output, target)
                RMSE = torch.sqrt(loss)
                if args.wandb is True:
                    wandb.log({
                        'eval_loss': loss,
                        'eval_RMSE': RMSE,
                        'epoch': epoch,
                        'glob_step': glob_step
                    })

                losses.update(loss.item(), bv.size(0))
                RMSEs.update(RMSE, bv.size(0))

                glob_step += 1
                loss_avg = losses.avg
                RMSE_avg = RMSEs.avg

                pbar.set_postfix({'loss': loss.item(), 'RMSE': RMSE.item()})
                pbar.update(bv.size(0))

    return loss_avg, RMSE_avg


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='EIT-3D Reconstruction with CNN')
    parser.add_argument('--wandb', default=False, help='if true, using wandb show training process.')
    parser.add_argument('--ampl', default=1000, type=int, help='Data amplification during calculation')
    parser.add_argument('--ddl', default='DDL', help='specific conductance, dir in the data')
    parser.add_argument('--reshape', default=False, help='Truth is do not use data refactoring, use reshape function')
    parser.add_argument('--pretrained', dest='pretrained', default=False, action='store_true',
                        help='use pre-trained model')
    parser.add_argument('--val-percent', default=0.2, type=float, help='Split into train and validation partitions')
    parser.add_argument('--workers', default=4, type=int, metavar='N', help='number of data loading workers')
    parser.add_argument('--num-elements', default=123074, type=int,
                        help='number of FEM elements(the number of outputs)  123074')
    parser.add_argument('--epochs', default=30, type=int, metavar='N', help='number of total epochs to run')
    parser.add_argument('--start-epoch', default=1, type=int, metavar='N',
                        help='manual epoch number (useful on restarts)')
    parser.add_argument('--es-patience', default=10, type=int, help='early stopping patience')
    parser.add_argument('--batch-size', default=32, type=int, metavar='N', help='mini-batch size (default: 2)')
    parser.add_argument('--step-size', default=30, type=int, metavar='N', help='step_size (default: 10)')
    parser.add_argument('--learning-rate', '--lr', default=0.001, type=float, metavar='LR',
                        help='initial learning rate',
                        dest='lr')
    parser.add_argument('--Net', default='ResNet50', help='Choose the Network style')
    parser.add_argument('--optimizers', default='Adam', help='Specify the optimizer type.include:SGD & Adam')
    parser.add_argument('--schedulers', default='RLROP', help='Specify the scheduler type.include:StepLR & RLROP')
    # parser.add_argument('--amp', action='store_true', default=True, help='Use mixed precision')
    parser.add_argument('--momentum', default=0.9, type=float, metavar='M', help='momentum')
    parser.add_argument('--weight-decay', default=1e-4, type=float, metavar='W', help='weight decay',
                        dest='weight_decay')
    parser.add_argument('--resume', default='', type=str, metavar='PATH',
                        help='path to latest checkpoint (default: none)')
    parser.add_argument('--seed', default=19940806, type=int, help='seed for initializing training. ')
    parser.add_argument('--gpu', default=0, type=str, help='GPU id to use.')

    start_all = time.time()

    log_path = './output/Logs/'
    if not os.path.exists(log_path):
        os.makedirs(log_path)
    # Log file names are set according to program runtime
    log_file_name = log_path + 'log-' + time.strftime("%Y%m%d-%H%M%S", time.localtime()) + '.log'
    # Record normal print messages
    sys.stdout = Logger(log_file_name)
    # Logging traceback exceptions
    sys.stderr = Logger(log_file_name)

    data_pairs = [
        ('BV_30dB_reshape', 'corn_30dB_ResNet'),
        ('BV_40dB_reshape', 'corn_40dB_ResNet'),
        ('BV_50dB_reshape', 'corn_50dB_ResNet')
    ]

    datadir = r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\corn_dataset_0.04_gradient'

    start_all = time.time()

    for i, pair in enumerate(data_pairs, start=2):
        main_bv, main_style = pair
        print(f'Run the {i}th main corn ear')
        main(
            datadir=datadir,
            main_bv=main_bv,
            main_style=main_style
        )

    end_all = time.time()
    print('Process Time:', end_all - start_all)
