
import argparse
import os
import time
import datetime
import numpy
from custom_function.function import Logger
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import pandas
import sys

parser = argparse.ArgumentParser(description='EIT-3D Reconstruction with CNN')
parser.add_argument('--num', default=100, help='The number of test sampels')
parser.add_argument('--ampl', default=1000, type=int, help='Data amplification during calculation')
parser.add_argument('--reshape', default=False, help='Truth is do not use data refactoring, use reshape function')
parser.add_argument('--ddl', default='DDL', help='specific conductance, dir in the data')
parser.add_argument('--workers', default=4, type=int, metavar='N', help='number of data loading workers')
parser.add_argument('--batch-size', default=1, type=int, metavar='N', help='mini-batch size (default: 2)')
parser.add_argument('--print-freq', default=1, type=int, metavar='N', help='print frequency')
parser.add_argument('--gpu', default=0, type=str, help='GPU id to use.')

def preprocess(pil_img, is_target):
    img_ndarray = numpy.asarray(pil_img)
    if img_ndarray.ndim == 2 and not is_target:
        img_ndarray = img_ndarray[numpy.newaxis, ...]
    elif not is_target:
        img_ndarray = img_ndarray.transpose((2, 0, 1))
    return img_ndarray

def main(datadir, main_bv, main_style, main_savepath, main_modelpath, main_logpath):
    RMSEs = []
    losses = []
    times_invp = []

    args = parser.parse_args()
    save_time = datetime.datetime.now().strftime('%m-%d-%H-%M')
    if not os.path.exists(main_logpath):
        os.makedirs(main_logpath)
    log_file_name = main_logpath + 'print_results_' + time.strftime("%Y%m%d-%H%M%S", time.localtime()) + '.log'
    sys.stdout = Logger(log_file_name)
    sys.stderr = Logger(log_file_name)
    '''
        Setup the device(GPU or CPU)
    '''
    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu), '; The name of device:', torch.cuda.get_device_name(0))
        torch.cuda.set_device(args.gpu)
    else:
        if torch.cuda.is_available():
            torch.device("cuda")
            print('Using device GPU',
                  '; The number of device:', torch.cuda.device_count(),
                  '; The name of device:', torch.cuda.get_device_name())
        else:
            torch.device("cpu")
            print('Using device CPU')
    model = torch.load(main_modelpath)
    print('Name of model:', main_modelpath)
    model.cuda()
    # summary(model.cuda(), input_size=(1, 45, 48))
    criterion = nn.MSELoss().cuda()

    for i in range(args.num):
        m = i + 1
        '''
        Data loading
        '''
        bv_path = os.path.join(datadir, main_bv, str(m) + '.csv')
        ddl_path = os.path.join(datadir, args.ddl, str(m) + '_ddl' + '.csv')
        # print(bv_path)
        # print(ddl_path)
        bv = pandas.read_csv(bv_path, header=None)
        ddl = pandas.read_csv(ddl_path, header=None)
        bv = preprocess(bv, is_target=True)
        ddl = preprocess(ddl, is_target=True)
        bv = torch.as_tensor(bv.copy()).float().contiguous()
        ddl = torch.as_tensor(ddl.copy()).float().contiguous()

        model.eval()

        if args.reshape is True:
            bv = torch.reshape(bv, (-1, 1, 45, 48))  # reshape
            bv = bv.cuda(non_blocking=True).to(dtype=torch.float32)
        else:
            bv = bv.cuda(non_blocking=True).to(dtype=torch.float32).unsqueeze(0).unsqueeze(0)
        ddl = ddl.cuda(non_blocking=True).to(dtype=torch.float32).squeeze(-1).unsqueeze(0)
        start_one = time.time()

        output = model(bv)
        end_one = time.time()

        ddl = ddl*args.ampl
        loss = criterion(output, ddl)
        # RMSE
        RMSE_val = torch.sqrt(loss)

        output = output/args.ampl
        ddl = ddl/args.ampl
        save_output = output.data.cpu()
        save_path_output = os.path.join(main_savepath, 'Output_{}'.format(main_style))
        if not os.path.exists(save_path_output):
            os.makedirs(save_path_output)
        save_name_output = os.path.join(save_path_output, str(m) + '_ddl' + '.csv')
        numpy.savetxt(save_name_output, save_output, fmt='%.4f', delimiter=',')

        save_ddl = ddl.data.cpu()
        save_path_ddl = os.path.join(main_savepath, 'DDL_{}'.format(main_style))
        if not os.path.exists(save_path_ddl):
            os.makedirs(save_path_ddl)
        save_name_ddl = os.path.join(save_path_ddl, str(m) + '_ddl' + '.csv')
        numpy.savetxt(save_name_ddl, save_ddl, fmt='%.4f', delimiter=',')

        losses.append(loss.data.cpu().numpy())
        RMSEs.append(RMSE_val.data.cpu().numpy())
        time_invp = end_one - start_one
        times_invp.append(time_invp)

    print('losses:', losses)
    print('RMSEs:', RMSEs)
    print('InvpTimes:', times_invp)

    losses_mean = numpy.mean(losses)
    RMSEs_mean = numpy.mean(RMSEs)
    losses_std = numpy.std(losses)
    RMSEs_std = numpy.std(RMSEs)

    InvpTime_mean = numpy.mean(times_invp)
    InvpTime_std = numpy.std(times_invp)

    print('-------------------------------- RMSEs & Loss & Time -------------------------------')
    print('losses mean:', losses_mean, '; ', 'losses standard deviation:', losses_std)
    print('RMSEs mean:', RMSEs_mean, '; ', 'RMSEs standard deviation:', RMSEs_std)
    print('Inverse problem time:', InvpTime_mean, '(mean); ', InvpTime_std, '(std)')

    RMSEs_csv = pandas.DataFrame(data=RMSEs)
    save_path_RMSEs = os.path.join(main_savepath, 'Result')
    if not os.path.exists(save_path_RMSEs):
        os.makedirs(save_path_RMSEs)
    save_name_RMSEs = os.path.join(save_path_RMSEs, 'RMSEs_{}'.format(main_style) + '.csv')
    RMSEs_csv.to_csv(save_name_RMSEs, header=False, index=False)

    times_invp_xlsx = pandas.DataFrame(data=times_invp)
    save_path_Invtime = os.path.join(main_savepath)
    if not os.path.exists(save_path_Invtime):
        os.makedirs(save_path_Invtime)
    save_name_Invtime = os.path.join(save_path_Invtime, 'Inv-prob-time_{}'.format(main_style) + '.csv')
    # numpy.savetxt(save_name_output, times_invp_xlsx, fmt='%.4f', delimiter=',')
    times_invp_xlsx.to_csv(save_name_Invtime, header=False, index=False)


if __name__ == '__main__':
    start_all = time.time()
    # print('oneball 30dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_oneball_1W_95295_v-max4.5_test',
    #     main_bv='BV_30dB_reshape',
    #     main_style='oneball_30dB_UNet',
    #     main_savepath=r'output/predict_oneball_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_oneball4.5-1W_30dB_UNet_model_50epochs_0.0001lr_05-19-21-12.pth',
    #     main_logpath=r'output/predict_oneball_UNet/',
    # )
    #
    # print('oneball 50dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_oneball_1W_95295_v-max4.5_test',
    #     main_bv='BV_50dB_reshape',
    #     main_style='oneball_50dB_UNet',
    #     main_savepath=r'output/predict_oneball_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_oneball4.5-1W_50dB_UNet_model_50epochs_0.0001lr_05-19-22-45.pth',
    #     main_logpath=r'output/predict_oneball_UNet/',
    # )
    #
    # print('oneprism 30dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_oneprism_1W_95295_v-max4.5_test',
    #     main_bv='BV_30dB_reshape',
    #     main_style='oneprism_30dB_UNet',
    #     main_savepath=r'output/predict_oneprism_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_oneprism4.5-1W_30dB_UNet_model_50epochs_0.0001lr_05-19-14-50.pth',
    #     main_logpath=r'output/predict_oneprism_UNet/',
    # )
    #
    # print('oneprism 50dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_oneprism_1W_95295_v-max4.5_test',
    #     main_bv='BV_50dB_reshape',
    #     main_style='oneprism_50dB_UNet',
    #     main_savepath=r'output/predict_oneprism_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_oneprism4.5-1W_50dB_UNet_model_50epochs_0.0001lr_05-19-16-27.pth',
    #     main_logpath=r'output/predict_oneprism_UNet/',
    # )
    #
    # print('onecylinder 30dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_onecylindrical_1W_95295_v-max4.5_test',
    #     main_bv='BV_30dB_reshape',
    #     main_style='onecylinder_30dB_UNet',
    #     main_savepath=r'output/predict_onecylinder_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_onecylinder4.5-1W_30dB_UNet_model_50epochs_0.0001lr_05-19-18-02.pth',
    #     main_logpath=r'output/predict_onecylinder_UNet/',
    # )
    #
    # print('onecylinder 50dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_onecylindrical_1W_95295_v-max4.5_test',
    #     main_bv='BV_50dB_reshape',
    #     main_style='onecylinder_50dB_UNet',
    #     main_savepath=r'output/predict_onecylinder_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_onecylinder4.5-1W_50dB_UNet_model_50epochs_0.0001lr_05-19-19-37.pth',
    #     main_logpath=r'output/predict_onecylinder_UNet/',
    # )
    #
    # print('threeobj 30dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_threeobj_1W_95295_v-max_4.5_test',
    #     main_bv='BV_30dB_reshape',
    #     main_style='threeobj_30dB_UNet',
    #     main_savepath=r'output/predict_threeobj_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_threeobj4.5-1W_30dB_UNet_model_50epochs_0.0001lr_05-19-11-37.pth',
    #     main_logpath=r'output/predict_threeobj_UNet/',
    # )
    #
    # print('threeobj 50dB')
    # main(
    #     datadir=r'D:\Projects\3DEIT-matlab\eidors-v3.10-ng\eidors\Creating_3D_forward_problem\absolute_threeobj_1W_95295_v-max_4.5_test',
    #     main_bv='BV_50dB_reshape',
    #     main_style='threeobj_50dB_UNet',
    #     main_savepath=r'output/predict_threeobj_UNet',
    #     main_modelpath=r'E:\UNet\BestRMSE_threeobj4.5-1W_50dB_UNet_model_50epochs_0.0001lr_05-19-13-16.pth',
    #     main_logpath=r'output/predict_threeobj_UNet/',
    # )

    print('oneball None')
    main(
        datadir=r'Your address\absolute_oneball_1W_95295_v-max4.5_test',
        main_bv='BV_reshape',
        main_style='oneball_None_ResNet',
        main_savepath=r'output/predict_oneball_ResNet',
        main_modelpath=r'Your address\BestRMSE_oneball_None_ResNet_model_50epochs_0.0001lr_08-03-11-07.pth',
        main_logpath=r'output/predict_oneball_ResNet/',
    )

    print('onecylinder None')
    main(
        datadir=r'Your address\absolute_onecylindrical_1W_95295_v-max4.5_test',
        main_bv='BV_reshape',
        main_style='onecylinder_None_ResNet',
        main_savepath=r'output/predict_onecylinder_ResNet',
        main_modelpath=r'Your address\BestRMSE_onecylindrical_None_ResNet_model_50epochs_0.0001lr_08-03-11-56.pth',
        main_logpath=r'output/predict_onecylinder_ResNet/',
    )

    print('oneprism None')
    main(
        datadir=r'Your address\absolute_oneprism_1W_95295_v-max4.5_test',
        main_bv='BV_reshape',
        main_style='oneprism_None_ResNet',
        main_savepath=r'output/predict_oneprism_ResNet',
        main_modelpath=r'Your address\BestRMSE_oneprism_None_ResNet_model_50epochs_0.0001lr_08-03-12-46.pth',
        main_logpath=r'output/predict_oneprism_ResNet/',
    )

    print('threeobj None')
    main(
        datadir=r'Your address\absolute_threeobj_1W_95295_v-max_4.5_test',
        main_bv='BV_reshape',
        main_style='threeobj_None_ResNet',
        main_savepath=r'output/predict_threeobj_ResNet',
        main_modelpath=r'Your address\BestRMSE_threeobj_None_ResNet_model_50epochs_0.0001lr_08-03-13-34.pth',
        main_logpath=r'output/predict_threeobj_ResNet/',
    )


    end_all = time.time()
    print('All Process Time:', end_all - start_all)
