import os
import numpy as np
import torch
import time
from openpyxl import load_workbook
import matplotlib.pyplot as plt
from tkinter import filedialog
import tkinter as tk
import pandas as pd
import argparse
import sys
from datetime import datetime
from tqdm import tqdm


# Prompt the user to select multiple files
def choose_files(title="Select Excel file", filedir=r"Enter your address"):
    filepaths = filedialog.askopenfilenames(title=title, filetypes=[("Excel files", "*.xlsx")], initialdir=filedir)
    return filepaths


# Prompt the user to select a save directory
def choose_save_directory(title="Select Save Directory"):
    directory = filedialog.askdirectory(title=title)
    return directory


def choose_model_file(title="Select Model", filetypes=None, modeldir=r"Enter your address"):
    root = tk.Tk()
    root.withdraw()
    filepath = filedialog.askopenfilename(title=title, filetypes=filetypes, initialdir=modeldir)
    return filepath


# Extract filenames from file paths and remove file extensions
def simplify_filename(filepath):
    base_name = os.path.basename(filepath)
    name_without_ext = os.path.splitext(base_name)[0]
    return name_without_ext


# Reading Excel data using openpyxl
def read_excel_data_openpyxl(filepath):
    wb = load_workbook(filepath, data_only=True)
    ws = wb["EITCALDATA (V)"]

    X1 = [ws.cell(row=i, column=3).value for i in range(11, 2171)]
    X2 = [ws.cell(row=i, column=4).value for i in range(11, 2171)]

    return np.array(X1), np.array(X2)


# Plot and save data
def plot_and_save_data(X1, X2, save_path):
    fig, ax = plt.subplots(2, 1)

    # Adjust line width and font size
    ax[0].plot(X1, linewidth=0.3)
    ax[0].set_xlabel('Measurement Points', fontsize=6)
    ax[0].set_ylabel('Real Part/V', fontsize=6)
    ax[0].tick_params(axis='both', labelsize=6)
    ax[0].set_xlim([0, len(X1)])

    ax[1].plot(X2, linewidth=0.3)
    ax[1].set_xlabel('Measurement Points', fontsize=6)
    ax[1].set_ylabel('Imaginary Part/V', fontsize=6)
    ax[1].tick_params(axis='both', labelsize=6)
    ax[1].set_xlim([0, len(X2)])

    fig.savefig(save_path, dpi=300, bbox_inches='tight')
    plt.close(fig)


# Select device and load model
def choose_device():
    if torch.cuda.is_available():
        print('Training with the GPU', '; Number of equipment:', torch.cuda.device_count(), '; device name:', torch.cuda.get_device_name(0))
        device = torch.device("cuda:0")
    else:
        print('using CPU')
        device = torch.device("cpu")
    return device


# Load the trained model
def load_trained_model(modelpath, device):
    model = torch.load(modelpath)
    model.to(device)
    model.eval()
    return model

def predict_with_model(model, data, device):
    tensor_data = torch.tensor(data, dtype=torch.float32).unsqueeze(0).unsqueeze(0).to(device)
    with torch.no_grad():
        output = model(tensor_data)
    return output.cpu().numpy()


def setup_logging(save_directory):

    current_timestamp = datetime.now().strftime('%Y%m%d-%H%M%S')
    log_filename = os.path.join(save_directory, f"processing_and_prediction_{current_timestamp}.log")
    sys.stdout = open(log_filename, 'w')
    return sys.stdout


def combined_process_and_predict(args):
    try:
        filepaths = choose_files()
        print("File selection complete.")

        save_directory = choose_save_directory()
        if not save_directory:
            print("Cancel save")
            return

        log_file = setup_logging(save_directory)
        print("Logging setup is complete..")

        prediction_times = []

        for filepath in tqdm(filepaths, desc="Progress in processing documents", ncols=100):
            print("\n" + "=" * 80)
            print(f"Start processing documents: {filepath}")
            print("=" * 80 + "\n")

            try:
                X1, X2 = read_excel_data_openpyxl(filepath)
                print("Data reading complete..")
            except Exception as e:
                print(f"An error occurred in the data reading module: {str(e)}")
                continue

            if len(X1) != 2160:
                print(f"Warning: File {filepath} data length is not 2160. Current length is {len(X1)}. Skip this file.")
                continue

            try:
                reshaped_X1 = X1.reshape(45, 48)
                print("Data conversion complete.")
            except Exception as e:
                print(f"An error occurred in the data conversion module: {str(e)}")
                continue

            try:
                save_name = simplify_filename(filepath)
                save_path_csv = os.path.join(save_directory, "Reshape_" + save_name + ".csv")
                pd.DataFrame(reshaped_X1).to_csv(save_path_csv, index=False, header=False)
                print(f"The extracted data has been converted 45*48 and saved as CSV file:. {save_path_csv}")

                save_path_X1_csv = os.path.join(save_directory, save_name + ".csv")
                pd.DataFrame(X1).to_csv(save_path_X1_csv, index=False, header=False)
                print(f"The extracted data has been saved as a CSV file {save_path_X1_csv}")

                save_path_png = os.path.join(save_directory, "PNG_" + save_name + ".png")
                plot_and_save_data(X1, X2, save_path_png)
                print(f"The data image has been saved as a PNG file. {save_path_png}")
            except Exception as e:
                print(f"An error occurred in the Data Saving module. {str(e)}")
                continue

            try:
                device = choose_device()
                model = load_trained_model(args.modelpath, device)
                print("Model loading complete.")

                start_one = time.time()
                model_output = predict_with_model(model, reshaped_X1, device)
                end_one = time.time()
                prediction_time = end_one - start_one
                prediction_times.append(prediction_time)
                print(f"Model predictions completed . Time-consuming:{prediction_time}seconds")
                model_output = model_output / args.ampl
            except Exception as e:
                print(f"An error occurred in the Model Prediction module: {str(e)}")
                continue

            try:
                save_path_output_csv = os.path.join(save_directory, 'DDL_' + save_name + ".csv")
                pd.DataFrame(model_output.squeeze()).to_csv(save_path_output_csv, index=False, header=False)
                print(f"Model predictions have been saved as a CSV file: {save_path_output_csv}")
                print(f"The file {filepath} is processed.\n")
            except Exception as e:
                print(f"An error occurred in the Save Model Predictions module: {str(e)}")
                continue

        if prediction_times:
            mean_time = np.mean(prediction_times)
            std_time = np.std(prediction_times)
            print(f"\nAverage prediction time: {mean_time:.4f} seconds, standard deviation: {std_time:.4f} seconds")
        else:
            print("No files were successfully processed, so the mean and standard deviation of the prediction times could not be calculated.")


    except Exception as e:
        print(f"An unknown error occurred in the main program: {str(e)}")


if __name__ == "__main__":
    start_time = time.time()
    parser = argparse.ArgumentParser(description='Data processing and model prediction tools')
    parser.add_argument('--ampl', default=10, type=int, help='Amplification factor for adjusting model output')
    parser.add_argument('--modelpath', type=str, default=None, help='Paths for pre-training models')
    args = parser.parse_args()

    if args.modelpath is None:
        args.modelpath = choose_model_file(title="Selecting pre-trained models", filetypes=[("PyTorch Model", "*.pth")])
        if not args.modelpath:
            print("No model is selected and the program exits.")
            sys.exit(1)

    combined_process_and_predict(args)

    end_time = time.time()
    total_time = end_time - start_time
    print(f"\nThe program was completed and took a total of time:{total_time:.2f}seconds")

    sys.stdout.close()
